import geopandas as gpd
import pandas as pd

from ..abstract import AbstractNode
from ..models import CRSProjectionModel
from ..models import DataFramePivotFields


class ConverterDataFrame2GeoDataFrame(AbstractNode):
    """
    Module performing projection of
    pd.DataFrame -> gpd.GeoDataFrame with `geometry` columns
    """

    def __init__(
        self,
        pivot_fields: DataFramePivotFields,
        crs_projection: CRSProjectionModel,
    ):
        self.pivot_fields = pivot_fields
        self.crs_projection = crs_projection

    def fit(self, X, y=None):
        return self

    def __factory_geometry(self, X: pd.DataFrame):
        return gpd.points_from_xy(
            X[self.pivot_fields.projected_lon],
            X[self.pivot_fields.projected_lat],
        )

    def __factory_geodataframe(
        self,
        X: pd.DataFrame,
        geometry: gpd.GeoSeries,
    ):
        return gpd.GeoDataFrame(
            X,
            geometry=geometry,
            crs=self.crs_projection.target_crs,
        )

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        geometry = self.__factory_geometry(X)
        X = self.__factory_geodataframe(X=X, geometry=geometry)
        return X



from typing import List
import pandas as pd

from ..abstract import AbstractNode


class UniqueTimestampConstraint(AbstractNode):
    """
    Module validating constrain that timestamp must be unique
    """

    def __init__(
        self,
        source_datetime: str,
    ):
        self.source_datetime = source_datetime

    def fit(self, X, y=None):
        return self

    def __get_mask_of_duplicates(self, X: pd.DataFrame) -> pd.Series:
        return X.duplicated([self.source_datetime])

    def __has_timestamp_duplicate(self, duplicate_mask: pd.Series) -> bool:
        return duplicate_mask.any()

    def __get_duplicated_timestamps(
        self,
        X: pd.DataFrame,
        duplicate_mask: pd.Series,
    ) -> List[str]:
        return X.loc[duplicate_mask, self.source_datetime].tolist()

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        duplicate_mask = self.__get_mask_of_duplicates(X)
        if self.__has_timestamp_duplicate(duplicate_mask):
            timestamps = self.__get_duplicated_timestamps(X=X, duplicate_mask=duplicate_mask)
            message = f"Duplicated timestamps met: {timestamps}"
            raise ValueError(message)
        return X



import pandas as pd
from ..abstract import AbstractNode
from ..models import DataFramePivotFields


class PanderaValidator(AbstractNode):
    """
    Node performing validation of pandera schema
    """

    def __init__(self, pivot_fields: DataFramePivotFields):
        self.pivot_fields = pivot_fields.pandera_schema

    def fit(self, X: pd.DataFrame, y=None):
        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        return self.pivot_fields.validate(X)



import pandas as pd
from sklearn.pipeline import Pipeline

from ..abstract import AbstractPredictor
from ..models import DataFramePivotFields


pivot_fields = DataFramePivotFields()


class ActivityExtractionSession(AbstractPredictor):
    """
    Class organizing life-cycle of activity extraction from
    GPS data  into form of session.

    Limitations:
    * 1 ActivityExtractionSession ~ 1 Vehicle
    * All components must be part of ecosystem
        `gps_activity_extraction.factory` module
    """

    def __init__(
        self,
        preprocessing: Pipeline,
        fragmentation: Pipeline,
        clustering: Pipeline,
        classification: Pipeline = None,
    ):
        """
        Args:
        preprocessing: Preprocessing pipeline module
        fragmentation: Fragmentation pipeline module
        clustering: Clustering pipeline module
        classification: CLassification pipeline module
        """
        self.preprocessing = preprocessing
        self.fragmentation = fragmentation
        self.clustering = clustering
        self.classification = classification

    def fit(self, X, y=None):
        return self

    def __preprocess_inputs(self, X: pd.DataFrame) -> pd.DataFrame:
        return self.preprocessing.transform(X)

    def __execute_fragmention(self, X: pd.DataFrame) -> pd.DataFrame:
        target_col = pivot_fields.fragmentation_output
        X[target_col] = self.fragmentation.predict(X)
        return X

    def __execute_clustering(self, X: pd.DataFrame) -> pd.DataFrame:
        target_col = pivot_fields.clustering_output
        X[target_col] = self.clustering.fit_predict(X)
        return X

    def __execute_classification(self, X: pd.DataFrame) -> pd.DataFrame:
        if self.classification is not None:
            target_col = pivot_fields.classification_output
            X[target_col] = self.classification.predict(X)
        return X

    def predict(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Clusters input dataframe by adding new columns

        Args:
        X: source dataframe

        Returns:
        pd.DataFrame: expanded dataframe
        """
        X = self.__preprocess_inputs(X)
        X = self.__execute_fragmention(X)
        X = self.__execute_clustering(X)
        X = self.__execute_classification(X)
        return X

    def get_fragmentation_input(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Transforms data by the stage of entering to fragmentation module

        Args:
        X: source dataframe

        Returns:
        pd.DataFrame: transformed dataframe
        """
        X = self.__preprocess_inputs(X)
        return X

    def get_clustering_input(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Transforms data by the stage of entering to clustering module

        Args:
        X: source dataframe

        Returns:
        pd.DataFrame: transformed dataframe
        """
        X = self.__preprocess_inputs(X)
        X = self.__execute_fragmention(X)
        return X

    def get_classification_input(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Transforms data by the stage of entering to classification module

        Args:
        X: source dataframe

        Returns:
        pd.DataFrame: transformed dataframe
        """
        X = self.__preprocess_inputs(X)
        X = self.__execute_fragmention(X)
        X = self.__execute_clustering(X)
        return X



from typing import Union
import pandas as pd


from ...abstract import AbstractPredictor


class VelocityFragmentator(AbstractPredictor):
    """
    Module selecting clustering candidates based on velocity value
    """

    def __init__(
        self,
        source_velocity_column: str,
        max_velocity_hard_limit: Union[float, int],
    ):
        self.source_velocity_column = source_velocity_column
        self.max_velocity_hard_limit = max_velocity_hard_limit

    def fit(self, X, y=None):
        return self

    def predict(self, X: pd.DataFrame) -> pd.DataFrame:
        X = self._get_input_copy(X)
        y = X[self.source_velocity_column] <= self.max_velocity_hard_limit
        return y.values



from .cluster_aggregator import ClusterAggregationFactory
from .coverage_statistics import CoverageStatisticsFactory
from .join_validator import JoinValidatorFactory
from .preprocessing import PreprocessingFactory
from .spatial_joiner import SpatialJoinerFactory

__all__ = [
    "ClusterAggregationFactory",
    "CoverageStatisticsFactory",
    "PreprocessingFactory",
    "SpatialJoinerFactory",
    "JoinValidatorFactory",
]



from abc import ABC, abstractmethod
from typing import List, Union

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.pipeline import Pipeline


class AbstractNode(ABC, BaseEstimator, TransformerMixin):
    @abstractmethod
    def fit(self, X: pd.DataFrame, y=None):
        pass

    @abstractmethod
    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        pass

    def fit_transform(self, X: pd.DataFrame, y=None):
        self.fit(X, y)
        return self.transform(X)


class AbstractPredictor(ABC, BaseEstimator, TransformerMixin):
    @abstractmethod
    def fit(self, X: pd.DataFrame, y=None):
        pass

    @abstractmethod
    def predict(self, X: pd.DataFrame) -> Union[List[Union[int, float, str]], np.array]:
        pass

    def _get_input_copy(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        NOTE: predictiors don't guarantee idempotancy, that's why important to copy object first
        """
        return X.copy()

    def fit_predict(self, X: pd.DataFrame, y=None) -> Union[List[Union[int, float, str]], np.array]:
        self.fit(X, y)
        return self.predict(X)


class AbstractPipelineFactory(ABC):
    @staticmethod
    @abstractmethod
    def factory_pipeline(*args, **kwargs) -> Pipeline:
        pass



from typing import List
import pandas as pd

from ..abstract import AbstractNode


class SelectColumns(AbstractNode):
    """
    Module selectng fields from dataframe
    """

    def __init__(self, columns: List[str]):
        self.columns = columns

    def fit(self, X: pd.DataFrame, y=None):
        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        return X[self.columns]



from typing import List
import pandas as pd

from ..abstract import AbstractNode
from ..models import DefaultValues


_KEYS_DELIMITER = DefaultValues().pk_delimiter


class PrimaryKeyGenerator(AbstractNode):
    """
    Module constructing single primary key from composite key
    """

    def __init__(
        self,
        target_column: str,
        source_columns: List[str],
    ):
        self.target_column = target_column
        if not isinstance(source_columns, list):
            raise ValueError("source columns must be list instance")
        self.source_columns = source_columns

    def fit(self, X, y=None):
        return self

    def __get_composite_key_fields(self, X: pd.DataFrame):
        _primary_key = X.loc[:, self.source_columns].astype(str).copy()
        return _primary_key

    def __expand_delimiters(self, X: pd.DataFrame) -> pd.DataFrame:
        X.iloc[:, :-1] = X.iloc[:, :-1] + _KEYS_DELIMITER
        return X

    def __assign_composite_key(self, X: pd.DataFrame, composite_keys: pd.Series):
        X.loc[:, self.target_column] = composite_keys.T.sum()
        return X

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        composite_keys = self.__get_composite_key_fields(X)
        composite_keys = self.__expand_delimiters(composite_keys)
        X = self.__assign_composite_key(X, composite_keys)
        return X



import numpy as np
import pandas as pd
from ...abstract import AbstractPredictor
from ...models import DataFramePivotFields
from ...models import DefaultValues


pivot_fields = DataFramePivotFields()
defaults = DefaultValues()


class STCM(AbstractPredictor):
    """
    Spatio-temporal clustering model performing clustering
    based on coordinates and time-stamp
    """

    # SOURCE FIELDS
    _LAT_COL = pivot_fields.projected_lat
    _LON_COL = pivot_fields.projected_lon
    _UNIXTIME_COL = pivot_fields.computed_unixtime
    _FRAGM_FLAG_COL = pivot_fields.fragmentation_output
    # INTERNAL UTILITY FIELDS
    _TEMP_CLUSTER_ID = pivot_fields.clustering_output
    _PROX_DIST_COL = "proximity_distance"
    _SPATIAL_PROXIMITY_FLAG = "is_spatially_near"
    _VALID_CLUSTERS_ID = "validated_cluster_id"
    # DEFAULTS
    _NOISE_GPS_ID = defaults.noise_gps_cluster_id
    _CLUST_CANDIDS_FRAC = 1
    # CLUSTER AGGREGATES DEFAULTS
    _AGG_N_STATIONARY = "n_stationary"
    _AGG_N_TOTAL = "n_total"
    _AGG_UNIX_START = "unixtime_start"
    _AGG_UNIX_END = "unixtime_end"
    _AGG_DURATION = "duration"
    _AGG_FRAC_STAT_PTS = "frac_stationary_points"

    def __init__(
        self,
        eps: float,
        min_duration_sec: float,
    ):
        """
        Args:
        eps (float): Max distance span between 2 adjacent
            gps records to link together
        min_duration_sec (float): Minimum duration spent on site to form cluster
        """
        self.eps = eps
        self.min_duration_sec = min_duration_sec

    def __compute_adj_prox_flag(self, X: pd.DataFrame):
        _adj_shift = 1
        d_lat = X[self._LAT_COL] - X[self._LAT_COL].shift(_adj_shift)
        d_lon = X[self._LON_COL] - X[self._LON_COL].shift(_adj_shift)
        X[self._PROX_DIST_COL] = np.sqrt(d_lat**2 + d_lon**2)
        X[self._SPATIAL_PROXIMITY_FLAG] = X[self._PROX_DIST_COL] <= self.eps
        return X

    def __generate_cluster_candidates(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        NOTE: Intuition is change is based on any inverted flag change
        Function generates initial ids for potential cluster candidates
            based on adjacent flag distance and stationary point flags

        Args:
        X (pd.DataFrame): input gps

        Returns:
        pd.DataFrame: gps with expanded cluster candidate id
        """
        _adj_prox_flag_inv = ~X[self._SPATIAL_PROXIMITY_FLAG]
        _stationary_flag_inv = ~X[self._FRAGM_FLAG_COL]
        _change_flag = _adj_prox_flag_inv | _stationary_flag_inv
        X[self._TEMP_CLUSTER_ID] = np.cumsum(_change_flag)
        return X

    def __agg_clusters_candidates(self, X: pd.DataFrame) -> pd.DataFrame:
        group_cols = [self._TEMP_CLUSTER_ID]
        agg_parameters = {
            self._AGG_N_STATIONARY: (self._FRAGM_FLAG_COL, "sum"),
            self._AGG_N_TOTAL: (self._TEMP_CLUSTER_ID, "count"),
            self._AGG_UNIX_START: (self._UNIXTIME_COL, "first"),
            self._AGG_UNIX_END: (self._UNIXTIME_COL, "last"),
        }
        X_agg = X.groupby(group_cols).agg(**agg_parameters).reset_index()
        X_agg[self._AGG_DURATION] = X_agg[self._AGG_UNIX_END] - X_agg[self._AGG_UNIX_START]
        X_agg[self._AGG_FRAC_STAT_PTS] = X_agg[self._AGG_N_STATIONARY] / X_agg[self._AGG_N_TOTAL]
        return X_agg

    def __valid_clust_candidates(self, X: pd.DataFrame) -> pd.DataFrame:
        _time_constr = X[self._AGG_DURATION] >= self.min_duration_sec
        _stat_pts_constr = X[self._AGG_FRAC_STAT_PTS] >= self._CLUST_CANDIDS_FRAC
        _validation_mask = _time_constr & _stat_pts_constr
        X = X.loc[_validation_mask, :].reset_index(drop=True)
        X[self._VALID_CLUSTERS_ID] = X.index
        return X

    def __set_cluster_ids_to_input_gps(self, X_original: pd.DataFrame, cluster_ids: pd.DataFrame):
        _join_columns = [self._TEMP_CLUSTER_ID, self._VALID_CLUSTERS_ID]
        derived_cluster_ids = cluster_ids.loc[:, _join_columns]
        X_original = X_original.merge(derived_cluster_ids, how="left")
        _id_values = X_original.loc[:, self._VALID_CLUSTERS_ID]
        _id_values = _id_values.fillna(self._NOISE_GPS_ID)
        _id_values = _id_values.astype(int)
        X_original[self._VALID_CLUSTERS_ID] = _id_values
        return X_original

    def __get_cluster_ids_vector(self, X: pd.DataFrame) -> np.array:
        return X[self._VALID_CLUSTERS_ID].values

    def fit(self, X: pd.DataFrame, y=None):
        return self

    def __validate_time_sequence(self, X: pd.DataFrame):
        time_delta = X[self._UNIXTIME_COL] - X[self._UNIXTIME_COL].shift(1)
        time_delta = time_delta.fillna(method="bfill")
        if (time_delta < 0).any():
            raise ValueError("Input gps time stamps must be ordered over time-stamp")

    def predict(self, X: pd.DataFrame) -> np.array:
        X = self._get_input_copy(X)
        self.__validate_time_sequence(X)
        X = self.__compute_adj_prox_flag(X)
        X = self.__generate_cluster_candidates(X)
        clusters = self.__agg_clusters_candidates(X)
        clusters = self.__valid_clust_candidates(clusters)
        X = self.__set_cluster_ids_to_input_gps(X, clusters)
        cluster_ids_vec = self.__get_cluster_ids_vector(X)
        return cluster_ids_vec



import pandas as pd

from ..abstract import AbstractNode


class SorterGPS(AbstractNode):
    """
    Module ordering gps records over unixtime
    """

    def __init__(
        self,
        unixtime_column: str,
    ):
        self.unixtime_column = unixtime_column

    def fit(self, X, y=None):
        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        X.sort_values(by=[self.unixtime_column], inplace=True)
        X.reset_index(drop=True, inplace=True)
        return X



from .cluster_aggregator import ClusterAggregator
from .coverage_statistics import CoverageStatistics
from .spatial_join_validator import SpatialJoinValidator
from .spatial_joiner import SpatialJoiner


__all__ = [
    "ClusterAggregator",
    "CoverageStatistics",
    "SpatialJoiner",
    "SpatialJoinValidator",
]



from typing import Dict, List, Tuple
import pandas as pd

from ..abstract import AbstractNode
from ..models import DataFramePivotFields


class DefaultSchemaProjector(AbstractNode):
    """
    Module projecting source user schema to default schema
    NOTE: this step is needed to work properly on fields at validation step
    """

    def __init__(
        self,
        original_pivot_fields: DataFramePivotFields,
    ):
        self.original_pivot_fields = original_pivot_fields
        self.target_pivot_fields = DataFramePivotFields()

    def fit(self, X, y=None):
        return self

    def __get_fields(self, pivot_fields: DataFramePivotFields) -> List[str]:
        return list(pivot_fields.__dict__.values())

    def __get_projection_candidates(self) -> Dict[str, str]:
        source_column_names = self.__get_fields(self.original_pivot_fields)
        target_column_names = self.__get_fields(self.target_pivot_fields)
        return list(zip(target_column_names, source_column_names))

    def __filter_pivot_field_overlaps(self, column_mapping: List[Tuple[str]]):
        projection_candidates = {}
        for target_column, source_column in column_mapping:
            if target_column != source_column:
                projection_candidates[target_column] = source_column
        return projection_candidates

    def __project_fields(
        self,
        X: pd.DataFrame,
        projection_mapping: Dict[str, str],
    ):
        for target_column, source_column in projection_mapping.items():
            X[target_column] = X[source_column]
        return X

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        projection_mapping = self.__get_projection_candidates()
        projection_mapping = self.__filter_pivot_field_overlaps(projection_mapping)
        X = self.__project_fields(X, projection_mapping)
        return X
